from ..errors import (
    MissingRequiredOauthArguments,
    NoScopesProvided
)

class OauthSession:
    """A class for managing an oauth session URL.
    
    Upon instantiating this class, you must call the set_scopes()
    method to select the scopes that have been selected on the 
    developer portal.
    
    Kwargs:
    -   `client_id` = The client ID from the dev portal.
    -   `redirect_uri` = The redirect URI from the dev portal.
    -   `response_type` = The response type. This should always be "code".
    -   `state` = The optional return state. Used for sessions.
    """
    BASE: str = "https://wants-to.party/oauth/authorize?"
    def __init__(
            self,
            *,
            client_id: str,
            redirect_uri: str,  
            response_type: str = "code",
            state: str = None
        ) -> None:
        self.client_id: str = client_id
        self.scopes: dict = {}
        self.redirect_uri: str = redirect_uri
        self.response_type: str = response_type
        self.state: str = state
    
    @classmethod
    def set_scopes(
            self,
            *,
            user_read: bool = False,
            user_email: bool = False,
            files_read: bool = False,
            files_write: bool = False,
            files_delete: bool = False
        ):
        """A required function that sets the scope 
        arguments to be included in the oauth URL.

        Kwargs corresponding to their scope:
        -   user_read = user.read
        -   user_email = user.email
        -   files_read = files.read
        -   files_write = files.write
        -   files_delete = files.delete

        """
        self.scopes = locals()
    
    @classmethod
    def _gen_url_friendly_scopes(self) -> str:
        final = []
        if self.scopes.get("user_read"):
            final.append("user.read")
        if self.scopes.get("user_email"):
            final.append("user.email")
        if self.scopes.get("files_read"):
            final.append("files.read")
        if self.scopes.get("files_write"):
            final.append("files.write")
        if self.scopes.get("files_delete"):
            final.append("files.delete")
        
        return "+".join(final)

    @classmethod
    def _check_required_arguments_exist(self) -> None:
        if not self.scopes:
            raise NoScopesProvided(
                "No scopes selected. Call the set_scopes() function to " \
                "set the scopes."
            )
        if not all(
            (self.client_id,
            self.redirect_uri,
            self.response_type)
        ):
            raise MissingRequiredOauthArguments(
                "You must provide a client_id, redirect_uri, and response_type"
            )

    @classmethod
    def _assemble_url(self) -> str:
        self._check_required_arguments_exist()

        client_id = self.client_id
        scopes = self._gen_url_friendly_scopes()
        redirect_uri = self.redirect_uri
        response_type = self.response_type
        state = "&" + self.state if self.state else ""

        sep = "&"

        return self.BASE + client_id + sep + scopes + sep + \
            redirect_uri + response_type + state
    
    @property
    def generated_url(self) -> str:
        """The final URL that's been generated by
        the information provided."""
        return self._assemble_url()

    @property
    def scopes(self) -> dict:
        """The scopes you have selected."""
        return self.scopes



import io
import os
import pathlib
from typing import Union
from aiohttp import ClientSession, FormData

from ..checks import Size
from ..errors import _handle_errorcode

class WantsToParty:
    """The base class used for all non-blocking wants-to.party API interactions.
    
    Keyword arguments
    -----------
    api_key - Your API key. This can be found on your dashboard.

    subdomain - Your unique subdomain. This can also be found on your dashboard.
    """
    def __init__(
            self,
            *,
            api_key: str,
            subdomain: str
        ):
        super().__init__()
        self.api_key: str = api_key
        self.subdomain: str = subdomain
        
        self._BASE: str = f"https://{subdomain}.wants-to.party/upload"
    
    async def _post_file(
            self,
            file,
            filetype: str,
            *,
            from_bytes: bool,
            filename: str = None,
            max_bytes: Size = None
        ) -> str:
        headers = {"key": self.api_key}
        
        if from_bytes:
            if max_bytes:
                max_bytes._check_bytes(file.getbuffer().nbytes)

            if not filetype.endswith("."):
                filetype = "." + filetype

            form = FormData(quote_fields=False)
            filename = f"file{filetype}"
            form.add_field("file", file.getvalue(), filename=filename)
        else:
            if max_bytes:
                max_bytes._check_bytes(os.path.getsize(file))

            form = FormData()
            form.add_field("file", open(file, "rb"), filename=filename + filetype)
        
        async with ClientSession(headers=headers) as session:
            async with session.post(self._BASE, data=form) as resp:
                payload = await resp.json()

                if resp.status != 200:
                    return _handle_errorcode(resp.status, payload["error"])

                return payload["url"]
    
    async def upload_from_bytes(
            self, 
            file: io.BufferedIOBase,
            filetype: str,
            *,
            max_bytes: Size = None
        ) -> str:
        """Uploads a file-like object using the raw binary data
        provided.

        Returns the URL of the uploaded file.

        Attributes
        -----------
        file - The raw binary data associated with the file.
               Using `io.BytesIO` is recommended.
            
        filetype - The file type associated with the file
                   you're uploading. I.E., ".png".

        max_bytes - A maximum number of bytes uploadable. Use
        the Size class to specify the number of bytes. 
        """
        return await self._post_file(file, filetype, from_bytes=True, max_bytes=max_bytes)
    
    async def upload_from_file(
            self,
            file: str,
            *,
            max_bytes: Size = None
        ) -> str:
        """Uploads a local file using the file path provided.
        
        Returns the URL of the uploaded file.
        
        Attributes
        -----------
        file - The path to the file stored on your local disk.

        max_bytes - A maximum number of bytes uploadable. Use
        the Size class to specify the number of bytes. 
        """
        filedata = pathlib.Path(file)
        
        filename = filedata.stem
        filetype = filedata.suffix
        path = os.path.abspath(file)
        return await self._post_file(path, filetype, filename=filename, from_bytes=False, max_bytes=max_bytes)



from typing import Union

from .errors import MaxFileSizeExceeded

def validate_file_extension(extension: str):
    """Returns `True` if the file extension provided is allowed 
    on wants-to.party.
    
    Allowed file extensions are:
    png, jpg, gif, mp4, mp3, json, txt, js, html, css, py

    Note that different variations of the same filetype don't work.
    E.G., "jpeg" won't work."""
    if extension.startswith("."):
        extension = extension.replace(".", "")

    allowed = [
        "png",
        "jpg",
        "gif",
        "mp4",
        "mp3",
        "json",
        "txt",
        "js",
        "html",
        "css",
        "py"
    ]
    
    if extension in allowed:
        return True
    
    return False

class Size:
    """A class that defines a maximum file size for a file.
    
    This transforms B, KB & MB units into the no. of bytes specified.
    
    To specify the maximum number of bytes, use the kwargs.

    If the maximum file size is exceeded, the MaxFileSizeExceeded exception is raised.
    
    GB+ units not necessary due to wants-to.party's maximum allowed file size of 100MB."""
    def __init__(
        self,
        *,
        mb: float = None,
        kb: float = None,
        b:  float = None
        ):
        self.mb: Union[float, bool] = mb
        self.kb: Union[float, bool] = kb
        self.b:  Union[float, bool] = b

        self.exceeded: bool = False
    
    def _check_bytes(self, all_bytes: int) -> None:
        nbytes = 1024
        b = 0
        kb = 0
        mb = 0

        if self.mb:
            mb = self.mb * (nbytes ** 2)
        
        if self.kb:
            kb = self.kb * (nbytes ** 1)
        
        if self.b:
            b = self.b
        
        total_allowed_bytes = mb + kb + b

        if all_bytes > total_allowed_bytes: # file exceeded max allowed filesize
            self.exceeded = True
            raise MaxFileSizeExceeded("This file exceeded the maximum file size limit.")
